import { NG_VALUE_ACCESSOR } from '@angular/forms';
import { QueryOperatorDirective } from './query-operator.directive';
import { QueryFieldDirective } from './query-field.directive';
import { QuerySwitchGroupDirective } from './query-switch-group.directive';
import { QueryButtonGroupDirective } from './query-button-group.directive';
import { QueryInputDirective } from './query-input.directive';
import { QueryRemoveButtonDirective } from './query-remove-button.directive';
import { ChangeDetectorRef, Component, ContentChild, ContentChildren, forwardRef, Input, } from '@angular/core';
export const CONTROL_VALUE_ACCESSOR = {
    provide: NG_VALUE_ACCESSOR,
    useExisting: forwardRef(() => QueryBuilderComponent),
    multi: true
};
export class QueryBuilderComponent {
    constructor(changeDetectorRef) {
        this.changeDetectorRef = changeDetectorRef;
        this.defaultClassNames = {
            removeIcon: 'q-icon q-remove-icon',
            addIcon: 'q-icon q-add-icon',
            button: 'q-button',
            buttonGroup: 'q-button-group',
            switchGroup: 'q-switch-group',
            queryTree: 'q-tree',
            queryItem: 'q-item',
            queryRule: 'q-rule',
            queryRuleSet: 'q-ruleset',
            invalidRuleset: 'q-invalid-ruleset',
            emptyWarning: 'q-empty-warning',
            fieldControl: 'q-field-control',
            operatorControl: 'q-operator-control',
            inputControl: 'q-input-control'
        };
        this.defaultOperatorMap = {
            string: ['=', '!=', 'contains', 'like'],
            number: ['=', '!=', '>', '>=', '<', '<='],
            time: ['=', '!=', '>', '>=', '<', '<='],
            date: ['=', '!=', '>', '>=', '<', '<='],
            category: ['=', '!='],
            boolean: ['=']
        };
        this.allowRuleset = true;
        /**
         * @deprecated since version 0.2.0
         */
        this.data = { condition: 'and', rules: [] };
        this.config = { fields: {} };
        this.defaultTemplateTypes = [
            'string', 'number', 'time', 'date', 'category', 'boolean', 'multiselect'
        ];
        this.defaultEmptyList = [];
        this.inputContextCache = new Map();
        this.operatorContextCache = new Map();
        this.fieldContextCache = new Map();
        this.removeButtonContextCache = new Map();
    }
    ngOnInit() { }
    ngOnChanges(changes) {
        const config = this.config;
        const type = typeof config;
        if (type === 'object') {
            this.fields = Object.keys(config.fields).map((value) => {
                const field = config.fields[value];
                field.value = field.value || value;
                return field;
            });
            this.operatorsCache = {};
        }
        else {
            throw new Error(`Expected 'config' must be a valid object, got ${type} instead.`);
        }
    }
    get value() {
        return this.data;
    }
    set value(value) {
        // When component is initialized without a formControl, null is passed to value
        this.data = value;
        this.changeDetectorRef.markForCheck();
        if (this.onChangeCallback) {
            this.onChangeCallback(value);
        }
    }
    writeValue(obj) {
        this.value = obj;
    }
    registerOnChange(fn) {
        this.onChangeCallback = fn;
    }
    registerOnTouched(fn) {
        this.onTouchedCallback = fn;
    }
    setDisabledState(isDisabled) {
        this.disabled = isDisabled;
    }
    findTemplateForRule(rule) {
        const type = this.getInputType(rule.field, rule.operator);
        if (type) {
            const queryInput = this.findQueryInput(type);
            if (queryInput) {
                return queryInput.template;
            }
            else {
                if (this.defaultTemplateTypes.indexOf(type) === -1) {
                    console.warn(`Could not find template for field with type: ${type}`);
                }
                return null;
            }
        }
    }
    findQueryInput(type) {
        const templates = this.parentInputTemplates || this.inputTemplates;
        return templates.find((item) => item.queryInputType === type);
    }
    getOperators(field) {
        if (this.operatorsCache[field]) {
            return this.operatorsCache[field];
        }
        let operators = this.defaultEmptyList;
        if (this.config.getOperators) {
            operators = this.config.getOperators(field);
        }
        const fieldObject = this.config.fields[field];
        const type = fieldObject.type;
        if (fieldObject && fieldObject.operators) {
            operators = fieldObject.operators;
        }
        else if (type) {
            operators = (this.operatorMap && this.operatorMap[type]) || this.defaultOperatorMap[type] || this.defaultEmptyList;
            if (operators.length === 0) {
                console.warn(`No operators found for field '${field}' with type ${fieldObject.type}. ` +
                    `Please define an 'operators' property on the field or use the 'operatorMap' binding to fix this.`);
            }
        }
        else {
            console.warn(`No 'type' property found on field: '${field}'`);
        }
        if (fieldObject.options) {
            operators = operators.concat(['in', 'not in']);
        }
        if (fieldObject.nullable) {
            operators = operators.concat(['is null', 'is not null']);
        }
        // Cache reference to array object, so it won't be computed next time and trigger a rerender.
        this.operatorsCache[field] = operators;
        return operators;
    }
    getInputType(field, operator) {
        if (this.config.getInputType) {
            return this.config.getInputType(field, operator);
        }
        const type = this.config.fields[field].type;
        switch (operator) {
            case 'is null':
            case 'is not null':
                return null; // No displayed component
            case 'in':
            case 'not in':
                return 'multiselect';
            default:
                return type;
        }
    }
    getOptions(field) {
        if (this.config.getOptions) {
            return this.config.getOptions(field);
        }
        return this.config.fields[field].options || this.defaultEmptyList;
    }
    getClassName(id) {
        const cls = this.classNames ? this.classNames[id] : null;
        return cls != null ? cls : this.defaultClassNames[id];
    }
    getDefaultOperator(field) {
        if (field && field.defaultOperator !== undefined) {
            return this.getDefaultValue(field.defaultOperator);
        }
        else {
            const operators = this.getOperators(field.value)[0];
            if (operators && operators.length) {
                return operators[0];
            }
            else {
                console.warn(`No operators found for field '${field.value}'. ` +
                    `A 'defaultOperator' is also not specified on the field config. Operator value will default to null.`);
                return null;
            }
        }
    }
    addRule(parent) {
        parent = parent || this.data;
        if (this.config.addRule) {
            return this.config.addRule(parent);
        }
        else {
            const field = this.fields[0];
            parent.rules = parent.rules.concat([
                {
                    field: field.value,
                    operator: this.getDefaultOperator(field)
                }
            ]);
        }
    }
    removeRule(rule, parent) {
        parent = parent || this.data;
        if (this.config.removeRule) {
            this.config.removeRule(rule, parent);
        }
        else {
            parent.rules = parent.rules.filter((r) => r !== rule);
        }
        this.inputContextCache.delete(rule);
        this.operatorContextCache.delete(rule);
        this.fieldContextCache.delete(rule);
        this.removeButtonContextCache.delete(rule);
    }
    addRuleSet(parent) {
        parent = parent || this.data;
        if (this.config.addRuleSet) {
            this.config.addRuleSet(parent);
        }
        else {
            parent.rules = parent.rules.concat([{ condition: 'and', rules: [] }]);
        }
    }
    removeRuleSet(ruleset, parent) {
        ruleset = ruleset || this.data;
        parent = parent || this.parentData;
        if (this.config.removeRuleSet) {
            this.config.removeRuleSet(ruleset, parent);
        }
        else {
            parent.rules = parent.rules.filter((r) => r !== ruleset);
        }
    }
    changeField(fieldValue, rule) {
        const field = this.config.fields[fieldValue];
        if (field && field.defaultValue !== undefined) {
            rule.value = this.getDefaultValue(field.defaultValue);
        }
        else {
            delete rule.value;
        }
        rule.operator = this.getDefaultOperator(field);
        // Create new context objects so templates will automatically update
        this.inputContextCache.delete(rule);
        this.operatorContextCache.delete(rule);
        this.fieldContextCache.delete(rule);
        this.getInputContext(rule);
        this.getFieldContext(rule);
        this.getOperatorContext(rule);
    }
    getDefaultValue(defaultValue) {
        switch (typeof defaultValue) {
            case 'function':
                return defaultValue();
            default:
                return defaultValue;
        }
    }
    getOperatorTemplate() {
        const t = this.parentOperatorTemplate || this.operatorTemplate;
        return t ? t.template : null;
    }
    getFieldTemplate() {
        const t = this.parentFieldTemplate || this.fieldTemplate;
        return t ? t.template : null;
    }
    getButtonGroupTemplate() {
        const t = this.parentButtonGroupTemplate || this.buttonGroupTemplate;
        return t ? t.template : null;
    }
    getSwitchGroupTemplate() {
        const t = this.parentSwitchGroupTemplate || this.switchGroupTemplate;
        return t ? t.template : null;
    }
    getRemoveButtonTemplate() {
        const t = this.parentRemoveButtonTemplate || this.removeButtonTemplate;
        return t ? t.template : null;
    }
    getQueryItemClassName(local) {
        let cls = this.getClassName('queryItem');
        cls += ' ' + this.getClassName(local.ruleset ? 'queryRuleSet' : 'queryRule');
        if (local.invalid) {
            cls += ' ' + this.getClassName('invalidRuleset');
        }
        return cls;
    }
    getButtonGroupContext() {
        if (!this.buttonGroupContext) {
            this.buttonGroupContext = {
                addRule: this.addRule.bind(this),
                addRuleSet: this.addRuleSet.bind(this),
                removeRuleSet: this.removeRuleSet.bind(this),
                $implicit: this.data
            };
        }
        return this.buttonGroupContext;
    }
    getRemoveButtonContext(rule) {
        if (!this.removeButtonContextCache.has(rule)) {
            this.removeButtonContextCache.set(rule, {
                removeRule: this.removeRule.bind(this),
                $implicit: rule
            });
        }
        return this.removeButtonContextCache.get(rule);
    }
    getFieldContext(rule) {
        if (!this.fieldContextCache.has(rule)) {
            this.fieldContextCache.set(rule, {
                changeField: this.changeField.bind(this),
                fields: this.fields,
                $implicit: rule
            });
        }
        return this.fieldContextCache.get(rule);
    }
    getOperatorContext(rule) {
        if (!this.operatorContextCache.has(rule)) {
            this.operatorContextCache.set(rule, {
                operators: this.getOperators(rule.field),
                $implicit: rule
            });
        }
        return this.operatorContextCache.get(rule);
    }
    getInputContext(rule) {
        if (!this.inputContextCache.has(rule)) {
            this.inputContextCache.set(rule, {
                options: this.getOptions(rule.field),
                field: this.config.fields[rule.field],
                $implicit: rule
            });
        }
        return this.inputContextCache.get(rule);
    }
}
QueryBuilderComponent.decorators = [
    { type: Component, args: [{
                selector: 'query-builder',
                template: `
    <ng-container *ngIf="getButtonGroupTemplate() as template; else defaultButtonGroup">
      <div [ngClass]="getClassName('buttonGroup')">
        <ng-container *ngTemplateOutlet="template; context: getButtonGroupContext()"></ng-container>
      </div>
    </ng-container>

    <ng-template #defaultButtonGroup>
      <div [ngClass]="getClassName('buttonGroup')">
        <button (click)="addRule()" [ngClass]="getClassName('button')">
          <i [ngClass]="getClassName('addIcon')"></i> Rule
        </button>
        <button (click)="addRuleSet()" [ngClass]="getClassName('button')" *ngIf="allowRuleset">
          <i [ngClass]="getClassName('addIcon')"></i> Ruleset
        </button>
        <ng-container *ngIf="!!parentData">
          <button (click)="removeRuleSet()" [ngClass]="getClassName('button')" class="danger">
            <i [ngClass]="getClassName('removeIcon')"></i>
          </button>
        </ng-container>
      </div>
    </ng-template>

    <ng-container *ngIf="getSwitchGroupTemplate() as template; else defaultSwitchGroup">
      <ng-container *ngTemplateOutlet="template; context: {$implicit: data}"></ng-container>
    </ng-container>

    <ng-template #defaultSwitchGroup>
      <div [ngClass]="getClassName('switchGroup')" *ngIf="data">
        <input type="radio" [(ngModel)]="data.condition" value="and" #andOption/>
        <label (click)="data.condition=andOption.value">AND</label>
        <input type="radio" [(ngModel)]="data.condition" value="or" #orOption/>
        <label (click)="data.condition=orOption.value">OR</label>
      </div>
    </ng-template>

    <ul [ngClass]="getClassName('queryTree')" *ngIf="data && data.rules">
      <ng-container *ngFor="let rule of data.rules">
        <ng-container *ngIf="{ruleset: !!rule.rules, invalid: !config.allowEmptyRuleset && rule.rules && rule.rules.length === 0} as local">
          <li [ngClass]="getQueryItemClassName(local)">
            <ng-container *ngIf="!local.ruleset">

              <ng-container *ngIf="getRemoveButtonTemplate() as template; else defaultRemoveButton">
                <div [ngClass]="getClassName('buttonGroup')">
                  <ng-container *ngTemplateOutlet="template; context: getRemoveButtonContext(rule)"></ng-container>
                </div>
              </ng-container>

              <ng-template #defaultRemoveButton>
                <div [ngClass]="getClassName('buttonGroup')">
                  <button class="danger" [ngClass]="getClassName('button')" (click)="removeRule(rule, data)">
                    <i [ngClass]="getClassName('removeIcon')"></i>
                  </button>
                </div>
              </ng-template>

              <ng-container *ngIf="getFieldTemplate() as template; else defaultField">
                <ng-container *ngTemplateOutlet="template; context: getFieldContext(rule)"></ng-container>
              </ng-container>

              <ng-template #defaultField>
                <select [ngClass]="getClassName('fieldControl')" [(ngModel)]="rule.field" (ngModelChange)="changeField($event, rule)">
                  <option *ngFor="let field of fields" [ngValue]="field.value">
                    {{field.name}}
                  </option>
                </select>
              </ng-template>

              <ng-container *ngIf="getOperatorTemplate() as template; else defaultOperator">
                <ng-container *ngTemplateOutlet="template; context: getOperatorContext(rule)"></ng-container>
              </ng-container>

              <ng-template #defaultOperator>
                <select [ngClass]="getClassName('operatorControl')" [(ngModel)]="rule.operator">
                  <option *ngFor="let operator of getOperators(rule.field)" [ngValue]="operator">
                    {{operator}}
                  </option>
                </select>
              </ng-template>

              <ng-container *ngIf="findTemplateForRule(rule) as template; else defaultInput">
                <ng-container *ngTemplateOutlet="template; context: getInputContext(rule)"></ng-container>
              </ng-container>

              <ng-template #defaultInput>
                <ng-container [ngSwitch]="getInputType(rule.field, rule.operator)">
                  <input [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'string'" type="text">
                  <input [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'number'" type="number">
                  <input [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'date'" type="date">
                  <input [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'time'" type="time">
                  <select [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'category'">
                    <option *ngFor="let opt of getOptions(rule.field)" [ngValue]="opt.value">
                      {{opt.name}}
                    </option>
                  </select>
                  <ng-container *ngSwitchCase="'multiselect'">
                    <div style="margin-bottom: 8px"></div>
                    <select [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" multiple>
                      <option *ngFor="let opt of getOptions(rule.field)" [ngValue]="opt.value">
                        {{opt.name}}
                      </option>
                    </select>
                  </ng-container>
                  <input [ngClass]="getClassName('inputControl')" [(ngModel)]="rule.value" *ngSwitchCase="'boolean'" type="checkbox">
                </ng-container>
              </ng-template>

            </ng-container>
            <query-builder *ngIf="local.ruleset"
              [data]="rule"
              [parentInputTemplates]="inputTemplates"
              [parentOperatorTemplate]="operatorTemplate"
              [parentFieldTemplate]="fieldTemplate"
              [parentSwitchGroupTemplate]="switchGroupTemplate"
              [parentButtonGroupTemplate]="buttonGroupTemplate"
              [parentRemoveButtonTemplate]="removeButtonTemplate"
              [parentData]="data"
              [config]="config"
              [allowRuleset]="allowRuleset"
              [operatorMap]="operatorMap">
            </query-builder>
            <p [ngClass]="getClassName('emptyWarning')" *ngIf="local.invalid">A ruleset cannot be empty. Please add a rule or remove it all together.</p>
          </li>
        </ng-container>
      </ng-container>
    </ul>
  `,
                styles: [`
    ﻿:host{display:block}:host .q-icon{font-style:normal;font-size:12px}:host .q-remove-icon::before{content:'❌'}:host .q-add-icon::before{content:'➕'}:host .q-switch-group,:host .q-button-group{font-family:"Lucida Grande", Tahoma, Verdana, sans-serif;overflow:hidden}:host .q-button-group{float:right}:host .q-button{margin-left:8px;background-color:white}:host .q-input-control,:host .q-operator-control,:host .q-field-control{display:inline-block;min-height:32px;padding:5px 8px;color:#555;background-color:#fff;background-image:none;border:1px solid #ccc;border-radius:4px;box-sizing:border-box;width:auto}:host .q-switch-group label,:host .q-button{float:left;min-height:30px;color:rgba(0,0,0,0.6);font-size:14px;font-weight:normal;text-align:center;text-shadow:none;padding:2px 8px;border:1px solid rgba(0,0,0,0.2);box-sizing:border-box;-webkit-transition:all 0.1s ease-in-out;-moz-transition:all 0.1s ease-in-out;-ms-transition:all 0.1s ease-in-out;-o-transition:all 0.1s ease-in-out;transition:all 0.1s ease-in-out}:host .q-switch-group label:hover,:host .q-button:hover{cursor:pointer;background-color:#F0F0F0}:host .q-switch-group label.success,:host .q-button.success{color:#75BE47}:host .q-switch-group label.danger,:host .q-button.danger{color:#B3415D}:host .q-switch-group input{position:absolute;clip:rect(0, 0, 0, 0);height:1px;width:1px;border:0;overflow:hidden}:host .q-switch-group label{background-color:#e4e4e4;line-height:24px}:host .q-switch-group input:checked+label{border:1px solid #619ed7;background:white;color:#3176b3}:host .q-invalid-ruleset{border:1px solid rgba(179,65,93,0.5) !important;background:rgba(179,65,93,0.1) !important}:host .q-empty-warning{color:#8d252e;text-align:center}:host .q-tree{list-style:none;margin:4px 0 2px}:host .q-tree .q-item{position:relative;padding:4px 6px;margin-top:4px;border:1px solid #CCCCCC;-webkit-transition:all 0.1s ease-in-out;-moz-transition:all 0.1s ease-in-out;-ms-transition:all 0.1s ease-in-out;-o-transition:all 0.1s ease-in-out;transition:all 0.1s ease-in-out}:host .q-tree .q-item.q-ruleset{background:rgba(204,204,204,0.2)}:host .q-tree .q-item.q-ruleset:hover{border:1px solid rgba(97,158,215,0.5);background:rgba(97,158,215,0.1)}:host .q-tree .q-item.q-rule{background:white}:host .q-tree .q-item::before{top:-5px;border-width:0 0 2px 2px}:host .q-tree .q-item::after{border-width:0 0 0 2px;top:50%}:host .q-tree .q-item::before,:host .q-tree .q-item::after{content:'';left:-12px;border-color:#CCC;border-style:solid;width:9px;height:calc(50% + 6px);position:absolute}:host .q-tree .q-item:last-child::after{content:none}
  `],
                providers: [CONTROL_VALUE_ACCESSOR]
            },] },
];
/** @nocollapse */
QueryBuilderComponent.ctorParameters = () => [
    { type: ChangeDetectorRef, },
];
QueryBuilderComponent.propDecorators = {
    'allowRuleset': [{ type: Input },],
    'classNames': [{ type: Input },],
    'operatorMap': [{ type: Input },],
    'data': [{ type: Input },],
    'parentData': [{ type: Input },],
    'config': [{ type: Input },],
    'parentInputTemplates': [{ type: Input },],
    'parentOperatorTemplate': [{ type: Input },],
    'parentFieldTemplate': [{ type: Input },],
    'parentSwitchGroupTemplate': [{ type: Input },],
    'parentButtonGroupTemplate': [{ type: Input },],
    'parentRemoveButtonTemplate': [{ type: Input },],
    'buttonGroupTemplate': [{ type: ContentChild, args: [QueryButtonGroupDirective,] },],
    'switchGroupTemplate': [{ type: ContentChild, args: [QuerySwitchGroupDirective,] },],
    'fieldTemplate': [{ type: ContentChild, args: [QueryFieldDirective,] },],
    'operatorTemplate': [{ type: ContentChild, args: [QueryOperatorDirective,] },],
    'removeButtonTemplate': [{ type: ContentChild, args: [QueryRemoveButtonDirective,] },],
    'inputTemplates': [{ type: ContentChildren, args: [QueryInputDirective,] },],
};
//# sourceMappingURL=query-builder.component.js.map